"""Test
"""
import unittest
from doctest import DocTestSuite
from doctest import DocFileSuite

import sparc.db.report

def test_suite():
    return unittest.TestSuite((
        DocFileSuite('period.txt',
                     package=sparc.db.report),))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')


from sqlalchemy.ext.declarative import declarative_base, declared_attr

from sparc.db import ISqlAlchemySession

# Class to act as a Base for all ORM-based SQLAlchemy mappings
Base = declarative_base()

class sparcBaseMixin(object):
    
    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()
    
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __mapper_args__= {'always_refresh': True}
    
    def __repr__(self):
        try:
            _id = self.getId()
        except KeyError:
            _id = None
        return "<" + self.__class__.__name__ + "(id: '%s')>" % (str(_id))

class EngineRequired(object):
    """Decorator for callables that are dependent on a SQLite SQLAlchemy engine
    
    Args:
        *args: String SQLAlchemy engine name that is required.  String must match
               a valid SQLAlchemy engine name (see engine.get_bind() method from
               SQLAlchemy...http://docs.sqlalchemy.org/en/rel_0_9/orm/session_api.html?highlight=get_bind#sqlalchemy.orm.session.Session.get_bind)
    
    Usage:
        Define a class that is decorated with EngineRequired().
        
        >>> from sparc.db import EngineRequired
        >>> @EngineRequired('sqlite','oracle')
        ... class myTestClass(object):
        ...     def __init__(self, aSessionArg):
        ...         self.session = aSessionArg
        
        Create a session, make sure it provides the sparc.db.ISqlAlchemySession 
        interface.  If your app is configured, and included configure.zcml 
        from sparc.db then this is already done, otherwise you'll need to do it
        yourself via zope.interface.directlyProvides().  We'll grab our ready-made
        session for this example that uses a sqlite memory-only db.
        
        >>> from sparc.common import Configure
        >>> import sparc.db
        >>> Configure([sparc.db, (sparc.db, 'memory.zcml')])
        >>> from z3c.saconfig import named_scoped_session
        >>> Session = named_scoped_session('memory_session')
        >>> session = Session()
        
        >>> from sparc.db import ISqlAlchemySession
        >>> ISqlAlchemySession.providedBy(session)
        True
        
        Now we'll test object instantiation.
        
        >>> myObject = myTestClass(session)
        >>> ISqlAlchemySession.providedBy(myObject.session)
        True
        
        Ok, now test another implementation that only uses oracle
        
        >>> @EngineRequired('oracle')
        ... class myFailingClass(object):
        ...     def __init__(self, aSessionArg):
        ...         self.session = aSessionArg
        
        
        >>> try:
        ...     myObject = myFailingClass(session)
        ... except LookupError:
        ...    print True
        True
        
    """
    def __init__(self, *args):
        self._allowed_engines = args
    
    def __call__(self, classThatRequiresSpecificEngine):
        def checkEngine(*args, **kwargs):
            session = None
            for a in args:
                if ISqlAlchemySession.providedBy(a):
                    session = a
                    break
            if not session:
                for a in kwargs.values():
                    if ISqlAlchemySession.providedBy(a):
                        session = a
                        break
            if not session:
                raise LookupError("expected class to be initialized with "\
                                  "ISqlAlchemySession argument (a SQLAlchemy session that provides the ISqlAlchemySession interface).")
            engine = session.get_bind()
            if engine.name not in self._allowed_engines:
                raise LookupError('Implementation requires SQLAlchemy engine of dialect %s, got: %s', str(self._allowed_engines), engine.name)
            return classThatRequiresSpecificEngine(*args, **kwargs)
        return checkEngine



from datetime import datetime, timedelta
from zope.interface import implements
from zope.component.factory import Factory
from sparc.db.report.interfaces import ITrendReportPeriod

class trendReportPeriod(object):
    """Iterator for time splits between begin/end given a set period
    
    Usage:
        >>> import datetime
        >>> begin = datetime.datetime(2015,1,1)
        >>> end   = datetime.datetime(2015,1,10)
        >>> period = 'day'
        >>> myPeriodIterator = trendReportPeriod(begin=begin,end=end,period=period)
        >>> myPeriodIterator.period
        'day'
        >>> mySplits = list(myPeriodIterator)
        >>> mySplits[0]
        (datetime.datetime(2015,1,1), datetime.datetime(2015,1,2))
        >>> mySplits[1]
        (datetime.datetime(2015,1,2), datetime.datetime(2015,1,3))
        >>> mySplits[9]
        (datetime.datetime(2015,1,9), datetime.datetime(2015,1,10))
    
    """
    implements(ITrendReportPeriod)
        
    def __init__(self, **kwargs):
        self._begin, self._end, self._period, self._label, self._cache = None, None, None, None, {'__iter__':{}, '_steps': {}} # initialize
        if 'begin' in kwargs: self.begin = kwargs['begin']
        if 'end' in kwargs: self.end = kwargs['end']
        if 'period' in kwargs: self.period = kwargs['period']
    
    @property
    def begin(self):
        return getattr(self, '_begin', None)
    @begin.setter
    def begin(self, begin):
        if begin and self._end and begin >= self._end:
            raise ValueError("expected begin datetime to be smaller than end datetime")
        if begin:
            self._begin = begin
    
    @property
    def end(self):
        return getattr(self, '_end', None)
    @end.setter
    def end(self, end):
        if end and self._begin and end <= self._begin:
            raise ValueError("expected end datetime to be larger than begin datetime")
        if end:
            self._end = end
    
    @property
    def period(self):
        return getattr(self, '_period', None)
    @period.setter
    def period(self, period):
        _allowed_strings = ['second', 'minute', 'hour', 'day', 'week', 'month', 'year']
        if isinstance(period, timedelta):
            self._period = period
        else:
            if not period in _allowed_strings:
                raise LookupError("Expected period to be valid string in %s, got: %s", str(_allowed_strings), str(period))
            self._period = period
    
    @property
    def label(self):
        _label = self._label
        if not _label:
            # timedeltas use simple numeric incrementer starting a 1
            if isinstance(self.period, timedelta): # 1, 2, 3...
                _label = lambda start, stop: str(self._steps().index((start, stop)) + 1)
            else:
                if self.period == 'second': _format = '%H:%M:%S' # HH:MM:SS
                if self.period == 'minute': _format = '%H:%M' # HH:MM
                if self.period == 'hour':  _format = '%m - %b %d, %H:%M' # XX - Mon DD, HH:MM (XX is zero-padded month number)
                if self.period == 'day': _format = '%m - %b %d %Y' # XX - Mon DD YYYY (XX is zero-padded month number)
                if self.period == 'week': _format = '%w - %a, %b %d %Y' # X - WDay, Mon DD YYYY (X is 0-6 weekday number, 0 is Sun)
                if self.period == 'month': _format = '%m - %b %Y' # XX - Mon YYYY (XX is zero-padded month number)
                if self.period == 'year': _format = '%Y' # YYYY
                _label = lambda start, stop: start.strftime(_format)
            
        return _label
    
    @label.setter
    def label(self, label):
        """label should be a callable with footprint label(start,stop)"""
        label(datetime(2013,1,1), datetime(2014,1,1), 'month') # simple test...can label be called?
        self._label = label
    
    def _steps(self):
        if (self._begin, self._end, self.period) in self._cache['_steps']:
            return self._cache['_steps'][(self._begin, self._end, self.period)]
        _steps = []
        progress = self.begin
        while progress < self.end:
            _next = self._getNextStep(progress)
            _steps.append((progress, _next))
            progress = _next
        self._cache['_steps'][(self._begin, self._end, self.period)] = _steps
        return _steps
    
    def __len__(self):
        return len(list(self.__iter__()))
    
    def __iter__(self):
        if (self._begin, self._end, self.period) in self._cache['__iter__']:
            return iter(self._cache['__iter__'][(self._begin, self._end, self.period)])
        
        if not self.begin or not self.end or not self.period:
            return iter([])
        _iter = []
        for _begin, _end in self._steps():
            _label = self.label(_begin, _end)
            _iter.append((_begin, _end, _label))
        self._cache['__iter__'][(self._begin, self._end, self.period)] = _iter
        """
        progress = self.begin
        while progress < self.end:
            _next = self._getNextStep(progress)
            _label = self.label(progress, _next)
            _iter.append((progress, _next, _label))
            progress = _next
        self._cache[(self._begin, self._end, self.period)] = _iter
        """
        return iter(_iter)
    
    def _getNextStep(self, progress):
        if not self.begin or not self.end or not self.period:
            return None
        if isinstance(self.period, timedelta):
            _next = progress + self.period
            return _next if _next < self.end else self.end
        
        # normalize the _start based on progress.  This insures next step
        # will land on our desired period increments.
        _start = {'year':progress.year, 'month':progress.month, 'day':progress.day, 'hour':progress.hour, 'minute':progress.minute, 'second':progress.second}
        for _period in ['second', 'minute', 'hour', 'week', 'day', 'month', 'year']: # week is before day intentionally...order matters
            if _period == self.period:
                if self.period in ['second', 'minute', 'hour', 'day', 'week']:
                    _delta = timedelta(**{self.period + 's': 1})
                    if self.period == 'week':
                        _delta_days = 7 - (progress.weekday() % 7) # diff days between next Monday and progress
                        _delta = timedelta(**{'days': _delta_days})
                    _next = datetime(**_start) + _delta
                    return _next if _next < self.end else self.end
                else: # 'month', 'year' are a little tougher
                    if self.period == 'year':
                        _start['year'] += 1
                        _next = datetime(**_start)
                        return _next if _next < self.end else self.end
                    else: # month
                        _start['month'] += 1
                        if _start['month'] <= 12:
                            _next = datetime(**_start)
                        else:
                            _start['month'] = 1
                            _start['year'] += 1
                            _next = datetime(**_start)
                        return _next if _next < self.end else self.end
            else:
                if _period in ['second', 'minute', 'hour']:
                    del _start[_period]
                if _period in ['month','year']:
                    _start['day'] = 1;
                if _period in ['year']:
                    _start['month'] = 1;
            
trendReportPeriodFactory = Factory(trendReportPeriod)
        
            


from zope.interface import Interface, Attribute

class IReportDates(Interface):
    begin = Attribute("Beginning datetime object of report metrics")
    end = Attribute("Ending datetime object of report metrics")

class IReportPublishInfo(Interface):
    name = Attribute("Report name")
    date = Attribute("datetime object of when report was generated")

class IReportCard(IReportDates, IReportPublishInfo):
    """A basic report card
    
    A report card is intended to act as a basic container for reports.  It
    provides a mechanism to gather reports into a logical group.
    """

class IReport(IReportDates, IReportPublishInfo):
    """A basic report
    
    A report is a collection of metrics based on a set of series vs. categories.
    A basic report might look like this:
    
    Example:
                          CATEGORIES
        SERIES | a basic category | another category
        ============================================
        set 1  |        12        |       8
        set 2  |        23        |       16
        set 3  |        220       |       134
        set 4  |        58        |       789
        
    """
    def categories():
        """Return list of categories to be applied to each series (think x)"""
    def series():
        """Return list of individual series to be reported on (think y)"""
    def metric(series, category):
        """Return value for the given metric"""

class ICulmulativeReport(IReport):
    """Report whose unique data points are included exclusively within a single metric.
    
    Example:
                          CATEGORIES
        SERIES        | Internal | External      Sum       Percentage
        ===================================
        Restricted    |   12     |   8           20           1.6
        Confidential  |   23     |   16          39           3.1
        Internal      |   220    |   134         354          28.1
        Public        |   58     |   789         847          67.2
        
        Sum               313        947         1260
        Percentage        24.8       75.2                     100
        
        This report classifies a file based on data sensitivity and location.  
        Each file can only be counted towards a unique/single cell.  If
        we added all the cells together, it would give the total number
        of files included in the report.
    """
    def sum(series = '', category = ''):
        """Return a integer series, category or complete cumulative total number"""
    def percentage(precision = 0, series = '', category = ''):
        """Return float percentage of series/category vs. all other series/categories"""

class ITrendReport(IReport):
    """Report on a set of metrics measured in discrete periods over time
    
    period starts are inclusive, while stops are exclusive (for report)
    """
    period = Attribute("Callable implementing ITrendReportPeriod")

class ITrendReportPeriod(IReportDates):
    """Divide begin/end dates into iterable of tuples of period length
    
    If period is a string, then individual periods will begin/end on midnight 
    (except for bookends which are defined by begin/end properties of the 
    object).  If period is a timedelta, then begin will act as the baseline for 
    all increments which are period in length, except for the end bookend.
    """
    period = Attribute("Valid period string [second, minute, hour, day, week, month, year] or Python timedelta")
    label = Attribute("Callable that accepts (start, stop) and returns a String series (y-axis) label")
    def __len__():
        """Return integer number of increments for the current period setting"""
    def __iter__():
        """Return tuple with begin/end/label (datetime/datetime/string) for each available increment"""


"""Test pwsecurity.sc.cache.corp_3pc_tracker module.
"""
import unittest
from doctest import DocTestSuite

import logging, sys

def test_suite():
    return unittest.TestSuite((DocTestSuite('sparc.db.sql'),))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')


from setuptools import setup, find_packages
import os

version = '0.0.3'

setup(name='sparc.db',
      version=version,
      description="Database interaction utilities for the SPARC platform",
      long_description=open("README.md").read() + "\n" +
                       open("HISTORY.txt").read(),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
      ],
      keywords=['zca'],
      author='David Davis',
      author_email='davisd50@gmail.com',
      url='https://github.com/davisd50/sparc.db',
      download_url = '',
      license='MIT',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['sparc'],
      include_package_data=True,
      package_data = {
          '': ['*.zcml']
        },
      zip_safe=False,
      install_requires=[
          'setuptools',
          'SQLAlchemy',
          'z3c.saconfig',
          'sparc.common'
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
